;       68HC12 KERNEL FOR WHYP12
;       Copyright 1998 by Richard E. Haskell
;       CSE Dept., Oakland University
;       Rochester, Michigan 48309

*********************************************************
********  Change the following five addresses   *********
********  for a particular 68HC12 environment  *********
*********************************************************
REGBASE EQU     $0000           ;register base
RAMBASE EQU     $0800           ;ram base
EESTART EQU     $0D00           ;start address of EEPROM
EESTOP  EQU     $0FFF           ;stop address of EEPROM
WHYPBASE EQU    $8000           ;start of WHYP kernel

******** To make a turnkey system change START  *******
******** to the address of your main word.      *******
******** The default, MN1, will run WHYP.       *******
***********************************************************
START   EQU     MN1

******** To use SCI1 instead of SCI0 change $C1-$C7 *******
******** in the following five lines to $C9-$CF     *******
***********************************************************
SC0BDL  EQU     REGBASE+$C1     ;baud rate control
SC0CR1  EQU     REGBASE+$C2     ;SCI control reg 1
SC0CR2  EQU     REGBASE+$C3     ;SCI control reg 2
SC0SR   EQU     REGBASE+$C4     ;SCI status reg
SC0DR   EQU     REGBASE+$C7     ;SCI data reg

COPCTL  EQU     REGBASE+$16     ;COP control reg
EEPROT  EQU     REGBASE+$F1     ;EEPROM Block Protect reg
EEPROG  EQU     REGBASE+$F3     ;EEPROM programming control reg
RDRF    EQU     $20             ;SCSR mask
STACK   EQU     RAMBASE+$80    ;system (return) stack 
DATSTK  EQU     RAMBASE+$40     ;data stack 
******  WHYP System Variables  ******
SP0     EQU     RAMBASE         ;initial data stack pointer
RP0     EQU     RAMBASE+$02     ;initial return stack pointer
LAST    EQU     RAMBASE+$04     ;address called by CREATE

***********************************************
*       WHYP Kernel
***********************************************

	ORG     WHYPBASE       ;beginning of whyp kernel
MAIN    
	LDS     #STACK          ;initialize system (return) stack
	STS     RP0
	LDX     #DATSTK         ;initialize data stack
	STX     SP0
	CLR     EEPROT          ;enable EEPROM writes 
	CLR     COPCTL          ;disable COP watchdog
	BSR     INITSER         ;initialize sci
	JMP     START           ;default=MN1, change for turnkey
MN1     BSR     INWDY           ;input word to Y
	JSR     0,Y             ;execute subroutine
	LDAA    #6
	BSR     OUTPUT          ;send ACK
	BRA     MN1

;       INITIALIZE SCI
INITSER 
	CLR     SC0CR1           ;8 bit
	LDD     #$340C
	STAA    SC0BDL           ;9600 baud
	STAB    SC0CR2           ;enable tx & rx
NOOP    RTS

;       INPUT BYTE FROM SERIAL PORT
INCHAR  
	LDAA    SC0SR           ;check status
	ANDA    #RDRF           ;check rdrf
	BEQ     INCHAR          ;wait for char
	LDAA    SC0DR           ;get char in A
	RTS

;       OUTPUT BYTE TO SERIAL PORT
OUTPUT  
	TST     SC0SR
	BPL     OUTPUT           ;loop until tdre
	STAA    SC0DR            ;send A
	RTS
	
;       PUSH INPUT WORD ON STACK 
;       LOW BYTE SENT FIRST
;       ( w -- )
TPUSH   
	BSR     INCHAR          ;get low byte
	STAA    1,-X            ;push on data stack
	BSR     INCHAR          ;get high byte
	STAA    1,-X            ;push on data stack
	RTS

;       POP STACK TO SERIAL PORT 
;       UPLOAD HIGH BYTE FIRST
;       ( -- w )
TPOP    
	PSHD
	LDD     2,X+            ;pop top of stack into D
	BSR     OUTPUT          ;send high byte
	TBA
	BSR     OUTPUT          ;send low byte
	PULD
	RTS

;       INPUT WORD TO Y
;       LOW BYTE SENT FROM HOST FIRST
INWDY   
	BSR     INCHAR
	TAB                     ;B=low byte
	BSR     INCHAR          ;D=word
	XGDY                    ;Y=word
	RTS

;       INPUT WORD TO X
INWDX   
	BSR     INCHAR
	TAB                     ;B=low byte
	BSR     INCHAR          ;D=word
	XGDX                    ;X=word
	RTS

;       STORE 'WORD' AT ADDR 
;       ( addr word -- )
STOREW
	PSHX                    ;save X
	BSR     INWDX           ;X = addr
	BSR     INWDY           ;Y = word
	XGDY                    ;D = word
	STD     0,X             ;store it
	PULX                    ;restore X
	RTS

;       STORE 'COUNT' BYTES AT ADDR IN EEPROM
;       ( #bytes addr sequence_of_bytes -- )
TBLKST  
	BSR     INWDY           ;Y = count
	PSHY                    ;count on stack
	BSR     INWDY           ;Y = addr
	CPY     #EESTART
	BLO     TBS2            ;If in eeprom
	CPY     #EESTOP
	BHI     TBS2
TBS1    BSR     INCHAR          ;get next byte
	BSR     EESTA           ;store it in eeprom
	INY                     ;inc addr
	PULD
	DECB                    ;dec count
	PSHD                    ;save count
	BNE     TBS1            ;do all 'count' bytes
	PULY                    ;fix stack
	RTS
TBS2    BSR     INCHAR          ;else
	STAA    0,Y             ;save in ram
	INY                     ;inc addr
	PULD                    ;get count
	DECB                    ;dec count
	PSHD                    ;save count
	BNE     TBS2            ;do all 'count' bytes
	PULY                    ;fix stack
	RTS

;       EEPROM PROGRAMMING ROUTINES
;       STORE BYTE A AT ADDRESS Y
EESTA
	PSHA                    ;save byte
	LDAA    0,Y             ;if @addr != FF
	CMPA    #$FF
	BEQ     EEB1
	BSR     BYTEE           ;erase byte
EEB1    PULA                    ;get byte
	LDAB    #$02
	STAB    EEPROG          ;set EELAT bit
	STAA    0,Y             ;store data to EEPROM addr
	LDAB    #$03
	STAB    EEPROG           ;set EEPGM bit (EELAT=1)
	JSR     DLY10           ;10 msec delay
	LDAB    #$02
	STAB    EEPROG          ;turn off hi voltage 
	CLR     EEPROG          ;clear EELAT bit
	RTS

;       ERASE BYTE AT ADDRESS Y
BYTEE
	LDAB    #$16
	STAB    EEPROG          ;set to byte erase mode
	STAB    0,Y             ;write any data to addr to erase
	LDAB    #$17
	STAB    EEPROG          ;turn on high voltage
	JSR     DLY10           ;10 msec delay
	LDAB    #$16
	STAB    EEPROG          ;turn off hi voltage 
	CLR     EEPROG          ;clear EELAT bit
	RTS

;       STORE WORD D AT ADDRESS Y AND Y+1
EESTD
	PSHB
	BSR     EESTA
	INY
	PULA
	BSR     EESTA
	RTS


*************************************
*       WHYP WORDS
*************************************

;       10MS.DELAY ( -- )       ; 8 MHz clock
DLY10                           ; 4*N + 12 = 80000
	PSHX                    ; 2 cycles
	LDX     #19997          ; 2  "     N = 19997
DL1     DEX                     ; 1  "
	BNE     DL1             ; 3  "
	PULX                    ; 3  "
	RTS                     ; 5  "

;       EEC!  ( c addr -- )
EECST
	LDY     2,X+
	LDD     2,X+
	TBA
	JSR     EESTA
	RTS

;       EE!  ( n addr -- )
EEST
	LDY     2,X+
	LDD     2,X+
	JSR     EESTD
	RTS

;       ERASE.BULK      ( -- )
BULKE
	LDAB    #$06
	STAB    EEPROG          ;set to bulk erase mode
	STAB    EESTART         ;write any data to EEPROM 
	LDAB    #$07
	STAB    EEPROG          ;turn on high voltage
	BSR     DLY10           ;10 msec delay
	LDAB    #$06
	STAB    EEPROG          ;turn off high voltage
	CLR     EEPROG          ;clear EELAT bit
	RTS

***********************************************
* 68HC12        WHYP Stack Words
***********************************************

; DUP ( w -- w w )
; Duplicate the top stack item.
DUPP    
	MOVW 0,X,2,-X        
	RTS             

; SWAP ( w1 w2 -- w2 w1 )
; Exchange top two stack items.
SWAP    
	LDD 0,X         
	MOVW 2,X,0,X         
	STD 2,X         
	RTS             

; OVER ( w1 w2 -- w1 w2 w1 )
; Copy second stack item to top.
OVER    
	MOVW 2,X,2,-X         
	RTS             

; DROP ( w -- )
; Discard top stack item.
DROP    
	INX             
	INX             
	RTS             

;       ROT     ( n1 n2 n3 -- n2 n3 n1 )
ROT
	LDD     0,X             ;D = n3
	MOVW    4,X,0,X         ;n1 n2 n1
	MOVW    2,X,4,X         ;n2 n2 n1
	STD     2,X             ;n2 n3 n1
	RTS

;       -ROT    ( n1 n2 n3 -- n3 n1 n2 )
MROT
	LDD     0,X             ;D = n3
	MOVW    2,X,0,X         ;n1 n2 n2
	MOVW    4,X,2,X         ;n2 n1 n2
	STD     4,X             ;n3 n1 n2
	RTS

;       NIP     ( n1 n2 -- n2 )
NIP
	MOVW    2,X+,0,X
	RTS

;       TUCK    ( n1 n2 -- n2 n1 n2 )
TUCK
	MOVW    0,X,2,-X        ;n1 n2 n2
	MOVW    4,X,2,X         ;n1 n1 n2
	MOVW    0,X,4,X         ;n2 n1 n2
	RTS

;       2DROP   ( n1 n2 -- )
DROP2
	INX
	INX
	INX
	INX
	RTS

;       2DUP    ( n1 n2 -- n1 n2 n1 n2 )
DUP2
	MOVW    2,X,2,-X
	MOVW    2,X,2,-X
	RTS

;       2SWAP   ( a b c d -- c d a b )
SWAP2
	LDD     0,X             ;D = d
	MOVW    4,X,0,X         ;a b c b
	STD     4,X             ;a d c b
	LDD     2,X             ;D = c
	MOVW    6,X,2,X         ;a d a b
	STD     6,X             ;c d a b
	RTS

;       2OVER   ( a b c d -- a b c d a b )
OVER2
	MOVW    6,X,2,-X
	MOVW    6,X,2,-X
	RTS

;       FLIP    ( ab -- ba )    flip bytes
FLIP    
	LDD     0,X
	STAB    0,X             ;bb
	STAA    1,X             ;ba
	RTS

;       PICK    ( n1 -- n2 )    copy n1th element to top 
PICK
	LDD     2,X+            ;D = n1
	NOP
	LSLD
	MOVW    D,X,2,-X
	RTS

;       ROLL    ( n -- )
ROLL
	LDD     0,X             ;D = n
	NOP
	LSLD
	LEAY    2,X             ;Y -> top of stack
	LEAY    D,Y             ;Y -> n1th element
	LSRB                    ;B = n
	MOVW    0,Y,0,X         ;save n1th element
	TSTB
	BEQ     RL2             ;move all elements
RL1     MOVW    2,-Y,2,Y        ; up 2 bytes
	DECB
	BNE     RL1
RL2     MOVW    2,X+,0,Y        ;move 1st element
	RTS

;       ?DUP    ( n -- 0 | n n )
QDUP
	LDAA    0,X             ;if n <> 0
	ORAA    1,X
	BEQ     QD1
	MOVW    0,X,2,-X        ;DUP
QD1     RTS

;       SP@ ( -- a )
; Push the current data stack pointer.
SPAT    
	PSHX
	PULY                    ;Y = X = a
	STY     2,-X            ;push on data stack
	RTS             

;       SP! ( a -- )
; Set the data stack pointer.
SPSTO   
	LDX   0,X               ;X = a
	RTS             

;       >R ( w -- )             ( R:  -- w )
; Push the data stack to the return stack.
TOR 
	PULY                    ;save return addr    
	LDD     2,X+            ;D = w
	PSHD                    ;push D on ret stack
	PSHY                    ;restore return addr
	RTS             

;       R> ( -- w )     ( R: w -- )
; Pop the return stack to the data stack.
RFROM   
	PULY                    ;save return addr
	PULD                    ;D = ret stack val, w
	STD     2,-X            ;push w on data stack
	PSHY                    ;restore return addr             
	RTS             

;       R@    ( -- w )  ( R: w -- w )
; Copy top of return stack to the data stack.
RAT     
	TSY                     ;SP+1 -> Y
	LDD     2,Y             ;D = ret stack val, w
	STD     2,-X            ;push w on data stack
	RTS             

;       R>DROP          ( R: sys -- )
RFDROP
	PULY                    ;save return addr
	PULD                    ;pop sys
	PSHY                    ;restore return addr
	RTS

;       RP@  ( -- a )
; Push the current RP to the data stack.
RPAT    
	PULD                    ;save return addr   
	TFR     SP,Y            ;SP -> Y
	STY     2,-X            ;push Y to data stack
	PSHD                    ;restore return addr
	RTS             

;       RP!     ( a -- )
; Set the return stack pointer.
RPSTO
	PULD                    ;save return addr   
	LDY     2,X+            ;Y=a
	TFR     Y,SP            ;Y -> SP, SP = a
	PSHD                    ;restore return addr
	RTS             


***********************************************
*       68HC12  WHYP Memory Access Words
***********************************************

;       !     ( w a -- )
STORE   
	LDY     2,X+            ;Y=a, pop data stack in Y 
	MOVW    2,X+,0,Y        ;store w at a    
	RTS             

;       @     ( a -- w )
AT      
	LDY     0,X             ;Y = a
	MOVW    0,Y,0,X         ;w = @Y    
	RTS             

;       C!    ( c b -- )      
CSTOR   
	LDY     2,X+            ;Y=b 
	LDD     2,X+            ;D = c 
	STAB    0,Y             ;store c at b
	RTS             

;       C@    ( b -- c )
CAT     
	LDY     0,X             ;Y=b 
	CLRA            
	LDAB    0,Y        
	STD     0,X         
	RTS             

***********************************************
*       68HC12  WHYP Logic Words
***********************************************

; AND ( w w -- w )
; Bitwise AND.
ANDD    
	LDD     2,X+            ;D=w, pop data stack into D 
	ANDA    0,X
	STAA    0,X
	ANDB    1,X
	STAB    1,X
	RTS

; OR ( w w -- w )
; Bitwise inclusive OR.
ORR     
	LDD     2,X+            ;D=w
	ORAA    0,X
	STAA    0,X
	ORAB    1,X
	STAB    1,X
	RTS

; XOR ( w w -- w )
; Bitwise exclusive OR.
XORR    
	LDD     2,X+            ;D=w 
	EORA    0,X
	STAA    0,X
	EORB    1,X
	STAB    1,X
	RTS

;       INVERT  ( n -- 1's_compl_of_n )
INVERT  
	COM     0,X
	COM     1,X
	RTS

;       TRUE = -1 = $FFFF
TRUE
	LDD     #-1
	STD     2,-X
	RTS

;       FALSE = 0
FALSE
	LDD     #0
	STD     2,-X
	RTS

***********************************************
*    68HC12  WHYP Relational Operator Words
***********************************************

;       0< ( n -- t )
; Return true if n is negative.
ZLESS   
	LDY     #$FFFF      ;true
	LDD     0,X         ;if not negative
	BMI     ZL1                
	LDY     #0000       ;false
ZL1     STY     0,X                
	RTS                    

;       0= ( n -- t )
; Return true if n is zero
ZEQ
	LDY     #$FFFF              ;true
	LDD     0,X             ;if not zero
	BEQ     ZE1            
	LDY     #0000           ;false
ZE1     STY     0,X             ;push flag on stack    
	RTS                 

;       < ( n1 n2 -- t )
; Return true if n1 is less than n2.
LT   
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = n2
	CPD     0,X             ;if n2 <= n1
	BGT     LT1
	LDY     #0              ;set false
LT1     STY     0,X             ;set flag
	RTS             

;       = ( n1 n2 -- t )
; Return true if n1 is equal to n2.
EQ
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = n2
	CPD     0,X             ;if n2 <> n1
	BEQ     EQ1
	LDY     #0              ;set false
EQ1     STY     0,X             ;set flag
	RTS             

;       0> ( n -- t )
; Return true if n is greater than zero
ZGT
	LDY     #$FFFF          ;true
	LDD     0,X             ;if <= zero
	BGT     ZG1            
	LDY     #0000           ;false
ZG1     STY     0,X             ;push flag on stack    
	RTS                 

;       > ( n1 n2 -- t )
; Return true if n1 is greater than n2.
GT
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = n2
	CPD     0,X             ;if n2 >= n1
	BLT     GT1
	LDY     #0              ;set false
GT1     STY     0,X             ;set flag
	RTS             

;       >= ( n1 n2 -- t )
; Return true if n1 is greater than or equal to n2.
GTE
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = n2
	CPD     0,X             ;if n2 > n1
	BLE     GTE1
	LDY     #0              ;set false
GTE1    STY     0,X             ;set flag
	RTS             

;       <= ( n1 n2 -- t )
; Return true if n1 is less than or equal to n2.
LTE
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = n2
	CPD     0,X             ;if n2 < n1
	BGE     LTE1
	LDY     #0              ;set false
LTE1    STY     0,X             ;set flag
	RTS             

;       <> ( n1 n2 -- t )
; Return true if n1 is not equal to n2.
NEQ
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = n2
	CPD     0,X             ;if n2 = n1
	BNE     NEQ1
	LDY     #0              ;set false
NEQ1    STY     0,X             ;set flag
	RTS             

;       U< ( u1 u2 -- t )
; Return true if u1 is less than u2.
ULT
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = u2
	CPD     0,X             ;if u2 <= u1
	BHI     ULT1
	LDY     #0              ;set false
ULT1    STY     0,X             ;set flag
	RTS             

;       U> ( u1 u2 -- t )
; Return true if u1 is greater than u2.
UGT
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = u2
	CPD     0,X             ;if u2 >= u1
	BLO     UGT1
	LDY     #0              ;set false
UGT1    STY     0,X             ;set flag
	RTS             

;       U<= ( u1 u2 -- t )
; Return true if u1 is less than or equal to u2.
ULE
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = u2
	CPD     0,X             ;if u2 < u1
	BHS     ULE1
	LDY     #0              ;set false
ULE1    STY     0,X             ;set flag
	RTS             

;       U>= ( u1 u2 -- t )
; Return true if u1 is greater than or equal to u2.
UGE
	LDY     #$FFFF          ;true
	LDD     2,X+            ;D = u2
	CPD     0,X             ;if u2 > u1
	BLS     UGE1
	LDY     #0              ;set false
UGE1    STY     0,X             ;set flag
	RTS             

***********************************************
*       68HC12 WHYP Arithmetic Words
***********************************************

;       + ( X Y -- X+Y )
PLUS    
	LDD     2,X+
	ADDD    0,X
	STD     0,X
	RTS

;       - ( X Y -- X-Y )
MINUS   
	LDD     2,X
	SUBD    2,X+
	STD     0,X
	RTS

;       D+ ( d1 d2 -- d3 )
DPLUS   
	LDAA    7,X
	ADDA    3,X
	STAA    7,X
	LDAA    6,X
	ADCA    2,X
	STAA    6,X
	LDAA    5,X
	ADCA    1,X
	STAA    5,X
	LDAA    4,X
	ADCA    0,X
	STAA    4,X
	LEAX    4,X
	RTS

;       D- ( d1 d2 -- d3 )
DMINUS  
	LDAA    7,X
	SUBA    3,X
	STAA    7,X
	LDAA    6,X
	SBCA    2,X
	STAA    6,X
	LDAA    5,X
	SBCA    1,X
	STAA    5,X
	LDAA    4,X
	SBCA    0,X
	STAA    4,X
	LEAX    4,X
	RTS

;   UM*         ( u u -- ud )   16 x 16 = 32
;               Unsigned multiply. Return double product.
UMSTA   
	LDD     2,X+                
	LDY     2,X+
	EMUL
	STD     2,-X
	STY     2,-X
	RTS                     
	
*       M*         ( n n -- d )  16 x 16 = 32
*       Signed multiply. Return double product.
MSTAR
	LDD     2,X+
	LDY     2,X+
	EMULS
	STD     2,-X
	STY     2,-X
	RTS


*       *         ( n n -- n )   16 x 16 = 16
*       Signed multiply. Return single product. (UM* DROP)
STAR
	LDD     2,X+                
	LDY     2,X+
	EMUL
	STD     2,-X
	RTS


;       DUM*    ( ud un -- ud )   32 x 16 = 32
;       Unsigned multiply of double number by single number.
;       Returns a double unsigned number, pcL pcH.
;       A B x C = pH pL  (drop high 16 bits of product, ACH)
;       pL = BCL, pH = BCH + ACL
DUMST
	LDD     0,X             ;D = un (C)
	LDY     2,X             ;Y = udH (A)
	EMUL                    ;Y = ACH, D = ACL
	STD     2,X             ;save ACL
	LDD     0,X             ;D = un (C)
	LDY     4,X             ;Y = udL (B)
	EMUL                    ;Y = BCH, D = BCL
	STD     4,X             ;save pL = BCL
	TFR     Y,D             ;D = BCH
	ADDD    2,X             ;D = BCH+ACL = pH
	STD     2,X             ;save pH
	LEAX    2,X             ;fix data stack
	RTS


*       UM/MOD      ( udl udh un -- ur uq )   32/16 = 16:16
*       Unsigned divide of a double by a single. Return mod and quotient.
UMMOD
	PSHX
	LDY     2,X
	LDD     4,X
	LDX     0,X
	EDIV
	BCS     UM1             ;if div by 0
	BVC     UM2             ; or overflow
UM1     LDD     #$FFFF          ; rem = $FFFF
	LDY     #$FFFF          ; quot = $FFFF
UM2     PULX
	LEAX    2,X
	STD     2,X
	STY     0,X
	RTS


*       M/MOD      ( dl dh n -- r q )    32/16 = 16:16
*       Signed divide of a double by a single. Return mod and quotient.
MMOD
	PSHX
	LDY     2,X
	LDD     4,X
	LDX     0,X
	EDIVS
	BCS     MM1             ;if div by 0
	BVC     MM2             ; or overflow
MM1     LDD     #$FFFF          ; rem = $FFFF
	LDY     #$FFFF          ; quot = $FFFF
MM2     PULX
	LEAX    2,X
	STD     2,X
	STY     0,X
	RTS


*       /MOD      ( n n -- r q )    16/16 = 16:16
*       Signed divide of a single by a single. Return mod and quotient.
SMOD
	PSHX
	LDD     2,X
	LDX     0,X
	IDIVS
	TFR     X,Y
	BCS     SM1             ;if div by 0
	BVC     SM2             ; or overflow
SM1     LDD     #$FFFF          ; rem = $FFFF
	LDY     #$FFFF          ; quot = $FFFF
SM2     PULX
	STD     2,X
	STY     0,X
	RTS


*       /         ( n n --  q )    16/16 = 16
*       Signed divide of a single by a single. Return quotient.
SLASH
	PSHX
	LDD     2,X
	LDX     0,X
	IDIVS
	TFR     X,Y
	BCS     SL1             ;if div by 0
	BVC     SL2             ; or overflow
SL1     LDD     #$FFFF          ; rem = $FFFF
	LDY     #$FFFF          ; quot = $FFFF
SL2     PULX
	LEAX    2,X
	STY     0,X
	RTS


*       U/         ( u u --  q )    16/16 = 16
*       Unsigned divide of a single by a single. Return quotient.
USLASH
	PSHX
	LDD     2,X
	LDX     0,X
	IDIV
	TFR     X,Y
	PULX
	LEAX    2,X
	STY     0,X
	RTS


*       MOD      ( n n -- r )    16/16 = 16
*       Signed divide of a single by a single. Return remainder.
MOD
	PSHX
	LDD     2,X
	LDX     0,X
	IDIVS
	TFR     X,Y
	BCS     MD1             ;if div by 0
	BVC     MD2             ; or overflow
MD1     LDD     #$FFFF          ; rem = $FFFF
MD2     PULX
	LEAX    2,X
	STD     0,X
	RTS

*       */      ( n1 n2 n3 -- n1*n2/n3 )    star-slash
*       Signed multiply-divide  n1*n2 double
STARSL
	PSHX
	LDD     4,X
	LDY     2,X
	EMULS
	LDX     0,X
	EDIVS
	BCS     SS1             ;if div by 0
	BVC     SS2             ; or overflow
SS1     LDY     #$FFFF          ; quot = $FFFF
SS2     PULX
	LEAX    4,X
	STY     0,X
	RTS

; IDIV    ( num denom -- rem quot )
;               Integer divide 16 by 16
IDIVV   
	PSHX
	LDD     2,X
	LDX     0,X
	IDIV                    ;X=quot,D=rem
	TFR     X,Y
	BCC     ID1             ;div by 0
	LDD     #$FFFF          ;quot already $FFFF
ID1     PULX
	STD     2,X
	STY     0,X
	RTS

;   FDIV        ( num denom -- rem quot )
;               Fractional divide 16 by 16
FDIVV   
	PSHX
	LDD     2,X
	LDX     0,X
	FDIV                    ;X=quot,D=rem
	TFR     X,Y
	BCS     FD1             ;if div by 0
	BVC     FD2             ; or numer > denom
FD1     LDD     #$FFFF          ; rem = $FFFF
	LDY     #$FFFF          ; quot = $FFFF
FD2     PULX
	STD     2,X
	STY     0,X
	RTS

;       MU/MOD  ( ud un -- urem udquot )
;       Unsigned divide of a double by a single.
;       Return remainder and double quotient.
MUMOD
	LDD     0,X
	STD     2,-X            ; dup un
	PSHX
	LDD     4,X             ;D = udH
	LDY     #0              ;0:udH / un
	LDX     0,X
	EDIV                    ;Y = quotH, D = remH
	BCC     MU1             ;if div by 0
	PULX
	LEAX    2,X
	LDD     #$FFFF          ; rem, quot = $FFFF
	STD     0,X
	STD     2,X
	STD     4,X
	RTS
MU1     PULX
	STY     2,X             ;quotH
	STD     4,X             ;remH
	LDY     4,X             ;Y = remH
	LDD     6,X             ;D = udL
	PSHX
	LDX     0,X             ;X = un
	EDIV                    ;Y = quotL, D = remL
	PULX
	STD     6,X             ;remL
	STY     4,X             ;quotL
	LEAX    2,X
	RTS



;       MIN     (n1 n2 -- min )
MIN
	LDY     2,X+            ;Y = n2
	CPY     0,X             ;if n2 < n1
	BGE     MIN1
	STY     0,X             ;leave n2 on stack
MIN1    RTS                     ;else leave n1

;       MAX     (n1 n2 -- min )
MAX
	LDY     2,X+            ;Y = n2
	CPY     0,X             ;if n2 > n1
	BLE     MX1
	STY     0,X             ;leave n2 on stack
MX1     RTS                     ;else leave n1

;       NEGATE  ( n -- -n )
NEGATE
	NEG     1,X             ;negate low byte
	BEQ     NG1             ;if not zero
	COM     0,X             ;complement high byte
	RTS
NG1     NEG     0,X             ;else, negate high byte
	RTS

;       ABS     ( n -- |n| )
ABS
	TST     0,X             ;if negative
	BPL     AB2             ;negate:
	NEG     1,X             ;negate low byte
	BEQ     AB1             ;if not zero
	COM     0,X             ;complement high byte
	RTS
AB1     NEG     0,X             ;else, negate high byte
AB2     RTS

;       1+      ( n -- n+1 )
ONEP
	LDD     0,X
	ADDD    #1
	STD     0,X
	RTS

;       2+      ( n -- n+2 )
TWOP
	LDD     0,X
	ADDD    #2
	STD     0,X
	RTS

;       1-      ( n -- n-1 )
ONEM
	LDD     0,X
	SUBD    #1
	STD     0,X
	RTS

;       2-      ( n -- n-2 )
TWOM
	LDD     0,X
	SUBD    #2
	STD     0,X
	RTS

;       2*      ( n -- 2*n )
TWOT
	ASL     1,X             ;arith shift left
	ROL     0,X
	RTS

;       2/      ( n -- n/2 )
TWOS
	ASR     0,X             ;arith shift right
	ROR     1,X
	RTS

;       U2/     ( u -- u/2 )
U2S
	LSR     0,X
	ROR     1,X             ;logic shift right
	RTS

;       +!      ( n addr -- )    add n to @addr
PSTORE
	LDY     2,X+            ;Y = addr
	LDD     0,Y             ;D = @addr
	ADDD    2,X+            ;add n to @addr
	STD     0,Y
	RTS

;       LSHIFT      ( n1 n2 -- n1 )  
;               Left shift bits of n1 n2 times
LSHIFT
	LDY     2,X+            ;Y = n2
LS1     ASL     1,X             ;arith shift left
	ROL     0,X             ; n2 bits
	DEY
	BNE     LS1
	RTS

;       RSHIFT      ( n1 n2 -- n1 )  
;               Right shift bits of n1 n2 times
RSHIFT
	LDY     2,X+            ;Y = n2
RS1     LSR     0,X             ;arith shift right
	ROR     1,X             ; n2 bits
	DEY
	BNE     RS1
	RTS

;       ?NEGATE ( n1 n2 -- n3 )  if n2 < 0, negate n1
QNEGATE
	TST     0,X             ;if n2 < 0
	BPL     QN2
	NEG     3,X             ;negate n1
	BEQ     QN1
	COM     2,X
	BRA     QN2
QN1     NEG     2,X
QN2     INX                     ;leave only n3
	INX
	RTS

;       DNEGATE ( d -- d )
DNEGATE
	NEG     3,X             ;dnegate
	BNE     DN1
	NEG     2,X
	BNE     DN2
	NEG     1,X
	BNE     DN3
	NEG     0,X
	RTS
DN1     COM     2,X
DN2     COM     1,X        
DN3     COM     0,X
	RTS


;       DABS    ( d -- |d| )
DABS
	TST     0,X             ;if d negative
	BPL     DA4
	NEG     3,X             ;dnegate
	BNE     DA1
	NEG     2,X
	BNE     DA2
	NEG     1,X
	BNE     DA3
	NEG     0,X
	RTS
DA1     COM     2,X
DA2     COM     1,X        
DA3     COM     0,X
DA4     RTS

;       S>D     ( n -- d )      sign extend
STOD
	LDD     #$FFFF          ;negative extend
	DEX
	DEX
	TST     2,X             ;if n positive
	BMI     SD1
	LDD     #$0000          ;extend zeros
SD1     STD     0,X             ;else extend 1's
	RTS

***********************************************
*       68HC12  WHYP Number I/O & Display
***********************************************

;       (LIT)   ( -- n )        
;               Runtime routine for single literals
LIT     
	PULY                    ;Y -> number
	LDD   0,Y               ;Push number
	STD   2,-X              ;on data stack
	INY                     ;jump over number
	INY
	PSHY
	RTS

;       (DLIT)  ( -- d )           
;               Runtime routine for double literals
DLIT    
	PULY                    ;Y -> number
	LDD   2,Y               ;Push number
	STD   2,-X              ;on data stack
	LDD   0,Y
	STD   2,-X
	INY                     ;jump over number
	INY
	INY
	INY
	PSHY
	RTS

;       . ( n -- )              ;print top of stack 
DOTT    
	LDAA  #1                ;opcode = 1
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send top of stack
	RTS

;       U. ( n -- )             ;print unsigned number
UDOTT   
	LDAA  #2                ;opcode = 2
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send top of stack
	RTS

;       D. ( d -- )             ;print double number
DDOTT   
	LDAA  #3                ;opcode = 3
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send double number
	JSR   TPOP
	RTS

;       UD. ( d -- )            ;print unsigned dnumber
UDDOTT  
	LDAA  #4                ;opcode = 4
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send double number
	JSR   TPOP
	RTS

;       EMIT ( c -- )           ;display ascii char 
EMIT    
	LDAA  #5                ;opcode = 5
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send top of stack
	RTS

;       .S      ( -- )  Display contents of stack
DOTS
	LDAA  #7
	JSR   OUTPUT            ;host code = 7
	STX   2,-X              ;save data stack ptr
	LDD   SP0
	SUBD  0,X               ;#bytes = SP0 - X
	ASRA
	RORB                    ;B = #stack items
	TBA
	JSR   OUTPUT            ;send #items
	TSTB
	BLE   DS2               ;while B > 0
	LDY   SP0               ; send bottom of stack 1st
DS1     PSHB
	LDD   2,-Y
	JSR   OUTPUT            ; send D
	TBA
	JSR   OUTPUT
	PULB
	DECB
	BNE   DS1               ; send all stack items
DS2     LEAX  2,X               ;fix data stack     
	RTS

;       .R      ( -- )  Display contents of return stack
DOTR
	LDAA  #8
	JSR   OUTPUT            ;host code = 8 (DOTS)
	TFR   SP,Y               ;Y = RP
	INY
	INY
	STY   2,-X              ;save return stack ptr
	LDD   RP0
	SUBD  0,X               ;#bytes = RP0 - RP
	ASRA
	RORB                    ;B = #stack items
	TBA
	JSR   OUTPUT            ;send #items
	TSTB
	BLE   DR2               ;while B > 0
	LDY   RP0               ; send bottom of stack 1st
DR1     PSHB
	LDD   2,-Y
	JSR   OUTPUT            ; send D
	TBA
	JSR   OUTPUT
	PULB
	DECB
	BNE   DR1               ; send all stack items
DR2     LEAX  2,X               ;fix data stack
	RTS

;       (.")    ( -- )  Run-time routine for ."
PDOTQ
	LDAA  #9
	JSR   OUTPUT            ;send 9
	PULY
	LDAA  0,Y
	JSR   OUTPUT            ;send count
	TAB                     ;B = count
DQ1     INY
	LDAA  0,Y               ;get next char
	JSR   OUTPUT            ;& send it
	DECB
	BNE   DQ1               ;do all chars
	INY
	PSHY                    ;skip over string
	RTS

;       S.FILE  ( start.addr end.addr -- )
;       send data for making s-record file
SFILE
	LDAA    #18
	JSR     OUTPUT          ;send code 18
SF1     JSR     INCHAR
	CMPA    #6
	BNE     SF1             ;wait for ACK
	LDD     2,X             ;send start.addr
	JSR     OUTPUT
	TBA
	JSR     OUTPUT
	LDD     0,X             ;send #bytes
	SUBD    2,X
	ADDD    #1
	JSR     OUTPUT
	TBA
	JSR     OUTPUT
	LDY     2,X             ;Y -> start.addr
SF2     LDAA    0,Y
	JSR     OUTPUT          ;send next byte
	INY
	CPY     0,X             ;repeat until 
	BLS     SF2             ; Y >= end.addr
	INX
	INX                     ;pop data stack
	INX
	INX
	RTS

;       (,)     ( n -- )
COMMA
	LDAA    #11                ; send code 11
	JSR     OUTPUT          ;get tdp
	JSR     INWDY           ;Y = tdp
	LDD     2,X+
	CPY     #$1000
	BLO     CMA1            ;if in eeprom
	CPY     #$1FFF
	BHI     CMA1
	JSR     EESTD           ; do eeprom store
	RTS                     ;else
CMA1    STD     0,Y             ; store tos at tdp in ram
	RTS

;       (C,)    ( c -- )
CCOMMA
	LDAA    #12             ; send code 12
	JSR     OUTPUT          ;get tdp
	JSR     INWDY           ;Y = tdp
	LDD     2,X+
	CPY     #$1000
	BLO     CCA1            ;if in eeprom
	CPY     #$1FFF
	BHI     CCA1
	TBA
	JSR     EESTA           ; do eeprom store
	RTS                     ;else
CCA1    STAB    0,Y             ;store char at tdp in ram
	RTS

;       CR      Carriage Return
CR
	LDAA  #$0D                 ; send 13
	JSR   OUTPUT
	RTS

;       .REG    ( -- )
;       Display registers
DOTREG  
	PSHY
	PSHX
	PSHB
	PSHA
	TPA
	PSHA                    ;push CCR
	BSR     DOTREGS
	PULA
	PULA
	PULB
	PULX
	PULY
	RTS

;       .REGS   ( -- )
;       Send register values to host
;       Pushed on stack by SWI
DOTREGS
	LDAA    #14
	JSR     OUTPUT          ;Code = 14
	TSY
	XGDY
	ADDD    #10             ;D = orig SP
	JSR     OUTPUT
	TBA
	JSR     OUTPUT          ;send SP
	TSY
	INY
	INY                     ;Y -> CCR
	LDAB    #9              ;9 bytes on stack
DRG1:   LDAA    0,Y
	JSR     OUTPUT          ;send 9 bytes on stack
	INY
	DECB
	BNE     DRG1
	LDY     2,-Y            ;get addr of next word
	LDD     0,Y             ; JSR D
	JSR     OUTPUT          ;send ^sub
	TBA
	JSR     OUTPUT
	RTS

***********************************************
*       68HC12 WHYP  -- CREATE ... DOES> 
***********************************************

;       (CREATE)        ( +++ )      Run time for CREATE
PCREATE
	LDAA    #15              ;code 15
	JSR     OUTPUT
PC1     JSR     INWDY           ;read word
	CPY     #6              ;if 6, exit
	BEQ     PC2
	JSR     0,Y             ;else, execute sub
	LDAA    #6
	JSR     OUTPUT          ;send ACK
	BRA     PC1
PC2     RTS

;       DOVAR   ( -- a )        (leave pfa on data stack)
				;Run time code for CREATE
DOVAR
	PULY                    ;get return addr (pfa)
	STY     2,-X            ;push on data stack
	RTS                     ;go to RTS of EXEC

;       (;CODE) ( -- )       Compile time code for DOES>
PSCODE
	PULD                    ;get addr of JSR DODOES
	LDY     LAST            ;replace DOVAR with DODOES 
	STD     0,Y             ;at addr in LAST
	RTS                     ;return up one level

;       DODOES  ( -- pfa )   Run time code for DOES>
DODOES
	PULY                    ;Y -> code following DOES>
	PULD                    ;D = pfa of defined word
	PSHY
	STD     2,-X            ;push pfa on data stack
	RTS                     ;return to code after DOES>

***********************************************
*       68HC12  WHYP Branching Words
***********************************************


;       EXIT  ( -- ), Exit the current word (subroutine)
EXIT    
	PULY            
	RTS             

;       EXECUTE ( ca -- )
EXECUTE 
	LDY     2,X+       
	JMP     0,Y         

;       I       ( -- n )        Index in DO loop
I
	TSY                       ;Y -> ret addr
	INY
	INY
	LDD     0,Y             ;D = index - (limit+$8000)
	ADDD    2,Y             ;D = index
	STD     2,-X             ;push index
	RTS

;       (DO)            Run time code for DO
PDO
	TSY
	LDD     0,Y             ;D = ret addr
	XGDY                    ;Y = ret addr, D = SP+1
	SUBD    #6
	XGDY                    ;D = ret addr, Y = SP+1-6
	STD     0,Y             ;put back ret addr
	TYS                     ;move SP to new ret addr
	INY
	INY                     ;Y -> return stack
	LDD     2,X             ;D = limit
	ADDD    #$8000          ;D = (limit + $8000)
	STD     2,Y             ;2nd element on ret stack
	STD     2,-X            ;save in temp storage
	LDD     2,X             ;D = index
	SUBD    0,X             ;D = index - (limit+$8000)
	STD     0,Y             ;1st element on ret stack
	PULY
	PSHY                    ;Y = next addr
	PSHY
	PULD                    ;D = next addr
	ADDD    0,Y             ;get xhere2
	TSY                     ;Y -> ret addr
	STD     6,Y             ;3rd element on ret stack
	INX
	INX                     ;pop temp storage
	INX
	INX
	INX
	INX                     ;pop index & limit
	PULY
	INY                     ;jump over displ
	INY
	PSHY
	RTS                     ;go to next sub

;       (LOOP)          Run time code for LOOP
PLOOP
	TSY   
	LDD     2,Y             ;inc top of ret stack
	ADDD    #1
	BVS     LP1             ;if no overflow
	STD     2,Y
	PULY                    ; keep looping
	PSHY                    ;Y -> displ
	PULD
	ADDD    0,Y             ;D -> sub after DO
	PSHD
	RTS                     ;jump there
LP1     LDD     0,Y             ;if overflow
	XGDY
	INY                     ;jump over displ
	INY
	ADDD    #6              ;rewind ret stack
	XGDY
	STD     0,Y
	TYS
	RTS                     ;& quit loop

;       LEAVE   ( -- )
LEAVE
	PULY                    ;jump
	TSY
	INY
	INY
	INY
	INY                     ;Y -> xhere2
	LDD     2,Y+            ;D = addr after LOOP
	TYS                     ;pop return stack
	PSHB
	PSHA                    ;push D
	RTS                     ;jump past LOOP

;       DONEXT  ( -- )          run time code for NEXT
DONEXT
	TSY
	LDD     0,Y             ;dec RP
	SUBD    #1
	STD     0,Y
	FCB     $26             ;BNE opcode
	
***********************************************
*       Misc. words             
***********************************************

;       MASK    8-bit byte
;       INPUT: B = bit no. 
;       OUTPUT; B = mask = 2^bit#
MASK
	PSHY                    ;save Y
	BSR     MSK1
	DB      1
	DB      2
	DB      4
	DB      8
	DB      16
	DB      32
	DB      64
	DB      128
MSK1    PULY                    ;addr of DB  1
	ABY
	LDAB    0,Y             ;get mask
	PULY                    ;restore Y
	RTS

;       HI      ( b# addr -- )
;       set bit number b# of byte at address addr to 1
HI     
	LDY     2,X+            ;Y = addr
	LDD     2,X+            ;B = b#
	BSR     MASK            ;B = mask
	STAB    1,-X            ;push B on data stack
	LDAA    0,Y             ;A = @Y
	ORAA    1,X+            ;OR with mask and pop mask
	STAA    0,Y             ;store back at addr
	RTS

;       LO      ( b# addr -- )
;       clear bit number b# of byte at address addr to 0
LO     
	LDY     2,X+            ;Y = addr
	LDD     2,X+            ;B = b#
	BSR     MASK            ;B = mask 
	COMB                    ;complement mask
	STAB    1,-X            ;push B on data stack
	LDAA    0,Y             ;A = @Y
	ANDA    1,X+            ;AND with mask and pop mask
	STAA    0,Y             ;store back at addr
	RTS

;       ?HI      ( b# addr -- f )
;       leave a true flag if bit number b# of byte at address addr is high
QHI     
	LDY     2,X+            ;Y = addr
	LDD     2,X+            ;B = b#
	BSR     MASK            ;b = mask 
	STAB    1,-X            ;push B on data stack
	LDAA    0,Y             ;A = @Y
	ANDA    1,X+            ;AND with mask and pop mask 
	BEQ     QH1             ;if not zero
	LDD     #$FFFF          ; leave true flag
	BRA     QH2
QH1     LDD     #$0000          ;else leave false flag
QH2     STD     2,-X            ;push flag on stack
	RTS

;       HERE    (  -- n )
HERE
	LDAA    #16
	JSR     OUTPUT          ;get tdp
	JSR     INWDY
	STY     2,-X            ;push on stack
	RTS

;       [']            Run-time of 'tick' in colon defn
BTICKB
	LDAA    #17              ;code 16
	JSR     OUTPUT
BT1     JSR     INWDY           ;read word
	CPY     #6              ;if 6, exit
	BEQ     BT2
	JSR     0,Y             ;else, execute sub
	LDAA    #6
	JSR     OUTPUT          ;send ACK
	BRA     BT1
BT2     RTS

;       DEPTH ( -- n )          return #items on stack
DEPTH
	STX     2,-X              ;save data stack ptr
	LDD     SP0
	SUBD    0,X               ;#bytes = SP0 - X
	ASRA
	RORB                      ;D = #stack items
	STD     0,X               ; if neg, underflow
	RTS

;       MOVE  ( from to cnt -- )
MOVE
	LDD     2,X+            ; D = count
	LDY     2,X+            ; Y = to address
	PSHX                    ; save X
	LDX     0,X             ; X = from address
MV1     MOVW    2,X+,2,Y+       ; move words from X to Y
	SUBD    #1              ; move cnt words
	BNE     MV1
	PULX                    ; restore X
	INX                     ; pop data stack
	INX
	RTS

;       CMOVE  ( ad1 ad2 cnt -- )
;       move cnt bytes up in memory from addr ad1 to ad2
CMOVE
	LDD     2,X+            ; D = count
	LDY     2,X+            ; Y = to address
	PSHX                    ; save X
	LDX     0,X             ; X = from address
CMV1    MOVB    1,X+,1,Y+       ; move bytes from X to Y
	SUBD    #1              ; move cnt bytes
	BNE     CMV1
	PULX                    ; restore X
	INX                     ; pop data stack
	INX
	RTS

;       CMOVE>  ( ad1 ad2 cnt -- )
;       move cnt bytes down in memory from addr ad1 to ad2
CMOVEG
	LDD     2,X+            ; D = count
	LDY     2,X+            ; Y = to address
	PSHX                    ; save X
	LDX     0,X             ; X = from address
	INY                     ; move bytes from 
	INX                     ;  right to left
CMVG1   MOVB    1,-X,1,-Y       ; move bytes from -X to -Y
	SUBD    #1              ; move cnt bytes
	BNE     CMVG1
	PULX                    ; restore X
	INX                     ; pop data stack
	INX
	RTS

**********************************
*       DEBUG WORDS
**********************************

*       HEX2ASC 
*       input:  A = hex value
*       output: A = ASCII value
HEX2ASC
	ANDA    #$0F
	CMPA    #9
	BLS     HA1
	ADDA    #$37
	RTS
HA1     ADDA    #$30
	RTS

*       EMIT blank space
BLANK
	MOVW    #$20,2,-X
	JSR     EMIT
	RTS

*       EMIT hex value in A
DOTXX
	PSHB
	PSHA
	LSRA
	LSRA
	LSRA
	LSRA
	BSR     HEX2ASC
	TAB
	CLRA
	STD     2,-X
	JSR     EMIT
	PULA
	BSR     HEX2ASC
	TAB
	CLRA
	STD     2,-X
	JSR     EMIT
	PULB
	RTS

*       EMIT 4 hex digit address in D
DOTADDR
	BSR     DOTXX
	TBA
	BSR     DOTXX
	RTS

*       DUMP16          Print out 1 line from 16 addresses
*       Input:  Y = addr
*       Output: Y = addr+16
DUMP16
	PSHD
	TFR     Y,D
	PSHY
	BSR     DOTADDR
	BSR     BLANK
	BSR     BLANK
	LDAB    #16
DP161   LDAA    1,Y+
	BSR     DOTXX
	BSR     BLANK
	DECB
	BNE     DP161
	BSR     BLANK
	LDAB    #16
	PULY
DP162   LDAA    1,Y+
	CMPA    #$20
	BHI     DP163
	LDAA    #$20
DP163   STAA    1,-X
	MOVB    #0,1,-X
	JSR     EMIT
	DECB
	BNE     DP162
	JSR     CR
	PULD
	RTS

*       HEAD16          ." dump header "
HEAD16
	JSR     PDOTQ
	DB    54       # of chars
	FCC   '       0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F '
	RTS

*       DUMP    ( addr cnt -- )
DUMP
	JSR     CR
	BSR     HEAD16
	JSR     CR
	LDD     2,X+            count
	LDY     2,X+            addr
	ADDD    #15
	LSRD                    B = # of 16-byte lines
	LSRD
	LSRD
	LSRD
DMP1    JSR     DUMP16
	DECB
	BNE     DMP1
	RTS

*       Send subroutine code to PC
*       SNDSUB  ( addr num_bytes -- )
SNDSUB
	LDD     2,X+            ;B = num_bytes
	LDY     2,X+            ;Y = addr
SSB1    LDAA    0,Y
	JSR     OUTPUT
	INY
	DECB
	BNE     SSB1
	RTS

*********************************************************
*       Fuzzy Control
*********************************************************

*       Fill Weight Arrays
*       fill.weights  ( xi memfunc_addr -- )
FILLWT
	LDY     2,X+            Y -> input membership functions
	LDD     2,X+            B = xi
	PSHX                    save X
	TFR     Y,X             X -> membership functions
	LDY     2,X             Y -> input weight table
	PSHB                    save xi
	LDD     4,X+            B = #mem functs; X->1st memb fnc
	PULA                    A = xi
FW1     MEM                     fuzzy membership grade
	DBNE    B,FW1
	PULX                    restore X
	RTS

*       Fire all rules
*       firerules ( in.out_addr rules_addr -- )
FIRERULES
	LDD     2,X+            D -> rules
	LDY     2,X+            Y -> in.out.array
	PSHX
	TFR     D,X             X -> rules
	INX
	INX                     X -> Wout addr
	PSHY                    save Y
	LDY     2,X+            Y -> Wout, X -> 1st rule
	LDAB    0,Y             B = #out.memb.fncs
FR0     CLR     1,X+             clear W array
	DBNE    B, FR0
	PULY                    Y -> in.out.array
	LDAA    #$FF
	REV                     rule evaluation
	PULX
	RTS

*       Find Output value
*       calc.output ( rules  cent_addr -- output_value )
FINDOUT
	LDD     2,X+            D -> cent
	LDY     2,X+            Y -> rules
	PSHX
	TFR     D,X             X -> cent array 
	LDY     2,Y             Y -> Warray
	LDAB    0,Y             B = #out.memb.fncs
	INY                     Y -> W array
	WAV
	EDIV                    Y = quotient
	TFR     Y,D             D = quotient  A = 0, B = output
	PULX
	STD     2,-X
	RTS

;       VHERE    (  -- n )
VHERE
	LDAA    #19             ;opcode = 19
	JSR     OUTPUT          ;get vdp
	JSR     INWDY
	STY     2,-X            ;push on stack
	RTS

;       ALLOT ( n -- )          ;allocate n bytes in code (tdp) space 
ALLOT    
	LDAA  #20               ;opcode = 20
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send top of stack
	RTS

;       VALLOT ( n -- )         ;allocate n bytes in ram (vdp) space 
VALLOT    
	LDAA  #21               ;opcode = 21
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send top of stack
	RTS

;       2!  ( d a -- )          store doulbe word
	LDY   2,X+              ;Y = a
	MOVW  2,X+,2,Y+         ;store dH
	MOVW  2,X+,0,Y          ;store dL
	RTS

;       2@  ( a -- d )          fetch double word
	LDY   2,X+              ;Y = a
	MOVW  2,Y,2,-X          ;push dL on data stack
	MOVW  0,Y,2,-X          ;push dH on data stack
	RTS

;       (")    ( -- )  Run-time routine for "
PQUOTE
	PULY                    ;get addr of cnt
	LDAB    0,Y
	CLRA                    ;D = count
	INY                     ;Y = addr
	STY     2,-X            ;push addr on data stack
	STD     2,-X            ;push len on data stack
	ABY                     ;add len to addr
	PSHY                    ;skip over string
	RTS

;       DARK ( -- )             ;clear the PC screen
DARK    
	LDAA  #22               ;opcode = 22
	JSR   OUTPUT            ;send to host
	RTS

;       AT ( col row -- )           ;set cursor on PC screen to col, row  
ATXY    
	LDAA  #23               ;opcode = 23
	JSR   OUTPUT            ;send to host
	JSR   TPOP              ;send row
	JSR   TPOP              ;send col
	RTS

;       SP0 ( -- addr )         ;get data stack pointer for empty stack
SPP0
	LDD   #SP0
	STD   2,-X
	RTS

;       RP0 ( -- addr )         ;get return stack pointer for empty stack
RPP0
	LDD   #RP0
	STD   2,-X
	RTS

;       B.LOOKUP  ( addr frac -- interp )   ;8-bit table lookup wrapper
BLOOKUP
	LDD   2,X+
	LDY   0,X
	TBL   0,Y
	TAB
	LDAA  #0
	STD   0,X
	RTS

;       W.LOOKUP  ( addr frac -- interp )   ;16-bit table lookup wrapper
WLOOKUP
	LDD   2,X+
	LDY   0,X
	ETBL  0,Y
	STD   0,X
	RTS

END     FCB     0

